import os
import pkg_resources

def startproject(project: str):
    """Generates a template project structure for an ML API project.

    Arguments:
        project::str- Project name
    
    Returns:
        None
    """
    try: 
        curr_dir = os.getcwd()
        root_dir = os.path.join(curr_dir, project)
        
        # creating the main project directory
        os.mkdir(root_dir) 

        # creating the files in root project directory
        f = os.path.join(root_dir, ".gitignore")
        with open(f, "w+") as gitignore:
            gitignore.write(" ")
        
        f = os.path.join(root_dir, "README.md")
        with open(f, "w+") as writefile:
            writefile.writelines(pkg_resources.resource_string('torchblaze', 'template_files/README.txt').decode('utf-8').split('\n'))

        f = os.path.join(root_dir, "requirements.txt")
        with open(f, "w+") as writefile:
            writefile.writelines(pkg_resources.resource_string('torchblaze', 'template_files/requirements.txt').decode('utf-8').split('\n'))
        
        f = os.path.join(root_dir, "app.py")
        with open(f, "w+") as writefile:
            writefile.writelines(pkg_resources.resource_string('torchblaze', 'template_files/app.py').decode('utf-8').split('\n'))
        
        f = os.path.join(root_dir, "tests.json")
        with open(f, "w+") as writefile:
            writefile.writelines(pkg_resources.resource_string('torchblaze', 'template_files/tests.txt').decode('utf-8').split('\n'))

        f = os.path.join(root_dir, "Procfile")
        with open(f, "w+") as writefile:
            writefile.writelines(pkg_resources.resource_string('torchblaze', 'template_files/procfile.txt').decode('utf-8').split('\n'))
        
        # creating the model directory and sub-dir/files
        model_dir = os.path.join(root_dir, 'model')
        os.mkdir(model_dir)
        os.mkdir(os.path.join(model_dir, 'data'))
        model_files = ['utils', 'model', 'train']
        
        for file in model_files:
            f = os.path.join(model_dir, file+'.py')
            with open(f, "w+") as writefile:
                    writefile.writelines(pkg_resources.resource_string('torchblaze', f'template_files/{file}.py').decode('utf-8').split('\n'))
    except:
        print(f"The directory '{project}' already exists. Kindly choose a different project name.")
    

if __name__ == "__main__":
    startproject('test_project')



import torch
import torch.nn as nn
import torch.nn.functional as F


class ParamsTooLargeException(Exception):
    pass


class ParamsTooSmallException(Exception):
    pass


class GradientAboveThresholdException(Exception):
    pass


class NaNParamsException(Exception):
    pass


class InfParamsException(Exception):
    pass


class GradientsUninitializedException(Exception):
    pass

class ParamsNotChangingException(Exception):
    pass

class DeviceNotCudaException(Exception):
    pass


def get_params(model):
    """Retrieves list of all the named parameters in a model.

    Arguments:
        model::torch.nn.Module- Ideally the deep learning model, but can be a set of layers/layer too.

    Returns:
        param_list::list- List of all the named parameters in the model.
    """
    return [(name, params) for name, params in model.named_parameters() if params.requires_grad]


def check_cuda(params):
    """Checks if the training device is of type CUDA or not.

    Arguments:
        params::torch.Tensor- The parameters associated with a model layer.
    
    Returns:
        None: Throws an exception if the training device is not CUDA-enabled.
    """ 
    try:
        assert params.device.type == "cuda"
    except AssertionError:
        raise DeviceNotCudaException("Training device is not of the type CUDA.")


def check_nan(name, params):
    """Tests for the presence of NaN values [torch.tensor(float("nan"))] in the given tensor of model parameter.

    Arguments:
        name::str- Name of the parameter
        params::torch.Tensor- Trainable named parameters associated with a layer

    Returns:
        None- Throws an exception in case any parameter is a NaN value.
    """
    try:
        assert not params.isnan().any()
    except AssertionError:
        raise NaNParamsException(f"\nNaN values found in the layer: {name}")


def check_infinite(name, params):
    """Tests for the presence of infinite values [torch.tensor(float("Inf"))] in the given tensor of model parameter.

    Arguments:
        name::str- Name of the parameter
        params::torch.Tensor- Trainable named parameters associated with a layer

    Returns:
        None- Throws an exception in case any parameter is a Inf value.
    """
    try:
        assert not params.isinf().any()
    except AssertionError:
        raise InfParamsException(
            f"\nInfinite values found in the layer: {name}")


def check_smaller(name, params, upper_limit=0):
    """Tests if the absolute value of any parameter exceeds a certain threshold.

    Arguments:
        name::str- Name of the parameter
        params::torch.Tensor- Trainable named parameters associated with a layer
        upper_limit::float- The threshold value every parameter should be smaller than in terms of its absolute value.

    Returns:
        None- Throws an exception in case any parameter is a exceeds threshold value.
    """
    try:
        assert params.abs().less(upper_limit).any()
    except AssertionError:
        raise ParamsTooLargeException(
            f"\nCertain parameters in layer '{name}' found to be greater than the threshold value = {upper_limit}.")


def check_greater(name, params, lower_limit=0):
    """Tests if the absolute value of any parameter falls below a certain threshold.

    Arguments:
        name::str- Name of the parameter
        params::torch.Tensor- Trainable named parameters associated with a layer
        lower_limit::float- The threshold value every parameter should be greater than in terms of its absolute value.

    Returns:
        None- Throws an exception in case any parameter is a NaN value.
    """
    try:
        assert params.abs().greater(lower_limit).any()
    except AssertionError:
        raise ParamsTooSmallException(
            f"\nCertain parameters in layer '{name}' found to be smaller than the threshold value = {lower_limit}.")


def check_gradient_smaller(name, params, grad_limit=1e3):
    """Tests if the absolute gradient value for any parameter exceed a certain threshold. Can be used to check for gradient explosion.

    Arguments:
        name::str- Name of the parameter
        params::torch.Tensor- Trainable named parameters associated with a layer
        grad_limit::float- A threshold value, such that, |params.grad| < grad_limit

    Returns:
        None- Throws an exception in case the gradient for any parameter exceeds the threshold value (grad_limit)
        OR
        None- Throws an exception in case the method is used without running the loss.backward() method for backprop.
    """
    grads = params.grad
    try:
        assert not (grads == None)
    except AssertionError:
        raise GradientsUninitializedException("\nModel gradients not initialized. Kindly run loss.backwards() to initialize gradients first.")

    try:
        assert not grads.abs().greater(grad_limit).any()
    except AssertionError:
        raise GradientAboveThresholdException(f"\nGradients (absolute) for certain parameters in layer '{name}' found to be greater than the threshold grad_limit value = {grad_limit}.")


def check_params_changing(params_list_old, params_list_new):
    """Tests if the parameters in the model/certain layer are changing after a training cycle.

    Arguments:
        name::str- Name of the parameter
        params_list_old::list- List of trainable model parameters BEFORE training cycle.
        params_list_new::list- List of trainable model parameters AFTER training cycle.

    Returns:
        None- Throws an exception in case the parameters are not changing 
    """
    try:
        for old, new in zip(params_list_old, params_list_new):
            _, params_old = old
            _, params_new = new
            assert params_old.equal(params_new)
        raise ParamsNotChangingException(f"\nModel parameters found to be NOT changing post training.")

    except AssertionError:
        print("Test for parameter change after training passed successfully.") 
        



def model_test(model, batch_x, batch_y, optim_fn,
               loss_fn=torch.nn.CrossEntropyLoss(), epochs=10, 
               test_gradient_smaller=True, test_greater=True,
               test_smaller=True, test_infinite=True, test_nan=True,
               test_cuda=False, test_params_changing=False, 
               upper_limit=1e1, lower_limit=1e-2, grad_limit=1e4):
    """Executes a suite of tests on the ML model.
    Set <test_name> = False if you want to exclude a certain test from the test suite.

    Arguments:
        model::nn.Module- The model that you want to test.

        batch_x::torch.Tensor- A single batch of data features to perform model checks

        batch_y::torch.Tensor- A single batch of data labels to perform model checks

        optim_fn::torch.optim- default=torch.optim.Adam, Optimizer algorithm to be used during model training 

        loss_fn- default=torch.nn.CrossEntropyLoss, Loss function to be used for model evaluation during training

        test_gradient_smaller::bool- default=True, Asserts if gradients exceed a certain threshold  

        test_greater::bool- default=True, Asserts if all parameters > threshold limit

        test_smaller::bool- default=True, Asserts if all parameters < threshold limit

        test_infinite::bool- default=True, Asserts that no parameters == infinite

        test_nan::bool- default=True, Asserts that no parameters == NaN

        test_params_changing::bool- Default=False, Asserts that all the parameters change/update after the training is complete

        test_cuda::bool- Default=False, Asserts that the model is training on a cuda-enabled GPU

        upper_limit::float- default=1e2, Absolute value of all parameters should be smaller than this threshold value

        lower_limit::float- default=1e-2, Absolute value of all parameters should be greater than this threshold value

        grad_limit::float- default=1e4, Absolute value of all gradients should be smaller than this threshold value
    """
    model.train() # putting the model in training mode

    model_params_old = get_params(model) # getting list of model parameters PRE training epoch
    
    if test_cuda: # needs to be checked only once
        check_cuda(model_params_old[0][1])

    for epoch in range(epochs): 
        optim_fn.zero_grad()
        output = model(batch_x)
        loss = loss_fn(output, batch_y)
        loss.backward()
        optim_fn.step()

        model_params = get_params(model) # getting list of model parameters POST training epoch

        for name, params in model_params: 
            if test_greater:
                check_greater(name, params, lower_limit=lower_limit)
            if test_smaller:
                check_smaller(name, params, upper_limit=upper_limit)
            if test_gradient_smaller:
                check_gradient_smaller(name, params, grad_limit=grad_limit)
            if test_nan:
                check_nan(name, params)
            if test_infinite:
                check_infinite(name, params)
        
        print(f"Epoch {epoch}: All tests passed successfully.")
    
    if test_params_changing:
        model_params_new = get_params(model)
        check_params_changing(model_params_old, model_params_new)   
            

if __name__ == "__main__":
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv1 = nn.Conv2d(3, 6, 5)
            self.pool = nn.MaxPool2d(2, 2)
            self.conv2 = nn.Conv2d(6, 16, 5)
            self.fc1 = nn.Linear(16 * 5 * 5, 120)
            self.fc2 = nn.Linear(120, 84)
            self.fc3 = nn.Linear(84, 10)

        def forward(self, x):
            x = self.pool(F.relu(self.conv1(x)))
            x = self.pool(F.relu(self.conv2(x)))
            x = x.view(-1, 16 * 5 * 5)
            x = F.relu(self.fc1(x))
            x = F.relu(self.fc2(x))
            x = self.fc3(x)
            return x

    net = Net()

    x = torch.rand(32,3,32,32)
    y = torch.randint(low=0, high=10, size=(32,))

    criterion = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(net.parameters(), lr=0.1, momentum=0.9)

    model_test(model=net, batch_x=x, batch_y=y, optim_fn=optimizer, loss_fn=criterion)

    



from flask import Flask, jsonify, request
from flask_restful import Api, Resource
import numpy as np
import traceback
import json
from model.model import Net
import torch
import cv2

app = Flask(__name__)
api = Api(app)

model = Net()
model.load_state_dict(torch.load('mnist_cnn.pt'))
model.eval()

# convert request_input dict to input accepted by model.
def parse_input(request_input):
    img = request_input.read()
    img = np.frombuffer(img, np.uint8)
    img = cv2.imdecode(img,cv2.IMREAD_COLOR)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    img = cv2.resize(img,(28,28))
    img = img.reshape((1,1,28,28))
    return torch.Tensor(img)


# convert model prediction to dict to return as JSON
def parse_prediction(prediction):
    prediction = prediction.argmax(dim=1, keepdim=True)
    return {'class':int(prediction[0][0])}

class MakePrediction(Resource):
    @staticmethod
    def post():
        if model:
            try:
                request_input = request.files['file']

                model_input = parse_input(request_input)

                prediction = model(model_input)

                model_output = parse_prediction(prediction)

                return jsonify(model_output)

            except:
                return jsonify({'trace': traceback.format_exc()})
        else:
            return jsonify({'trace': 'No model found'})

api.add_resource(MakePrediction, '/predict')


class dummy_post(Resource):
    @staticmethod
    def post():
        if model:
            try:
                request_input = request.get_json()

                return jsonify(request_input)

            except:
                return jsonify({'trace': traceback.format_exc()})
        else:
            return jsonify({'trace': 'No model found'})

api.add_resource(dummy_post, '/dummy_post')


class dummy_get(Resource):
    @staticmethod
    def get():
        if model:
            try:
                return str("dummy get request")

            except:
                return jsonify({'trace': traceback.format_exc()})
        else:
            return jsonify({'trace': 'No model found'})

api.add_resource(dummy_get, '/dummy_get')

if __name__ == '__main__':
    app.run(debug=True,host='0.0.0.0',port=8080)



import importlib
import os
from flask import url_for
import sys
import requests
import json
# Setting the System Path to current directory

curr_dir = os.getcwd()
sys.path.append(curr_dir)



def has_no_empty_params(rule):
    """Check whether the route rules contain any empty parameters or not

    Arguments:
        input takes the route which we want to check 
        
    Returns:
        Bool: True if the route doesn't contain any empty parameters, otherwise false.
    """
    
    defaults = (rule.defaults if rule.defaults is not None else ())
    arguments = (rule.arguments if rule.arguments is not None else ())
    return len(defaults) >= len(arguments)


def get_routes():
    """Get all the list of routes with the functions initialised in the app.py file

    Arguments:
        None

    Returns:
        list: returns the list of routes defined in app.py
    """
    
    # Loading the app.py file as a module using library importlib

    from app import app
    routes = []
    for rule in app.url_map.iter_rules():
        # Checking whether the rule has any empty params

        if has_no_empty_params(rule):
            # The list contains the tuple which comprises of route method,route path,route end point

            routes.append((rule.methods,str(rule),rule.endpoint))
    return routes


def tests(routes,baseurl):
    """It sends the request to the routes by taking test cases from tests.json"

    Arguments:
        input takes list of routes and the baseurl of the api

    Returns:
        None
    """
    # Loading tests.json file
    curr_dir=os.getcwd()
    f = os.path.join(curr_dir, "tests.json")
    with open(f, "r") as jsonfile:
        data=json.load(jsonfile)
        #print(data)
    for i in routes:
        
        # GET Method Testing
        if 'GET' in i[0]:
            #print("get")
            #print(baseurl)
            response=requests.get(baseurl+str(i[1]))
            status_code_get=response.status_code
            if(status_code_get==200):
                print("route",str(i[1]),"get successful")
            else:
                print(i[1],"failed with return status_code",status_code_get)
        # POST Method Implemntation
        elif 'POST' in i[0]:
            endpoint=str(i[2])
            if endpoint=='makeprediction':
                continue
            end_test_data=data[endpoint]
            for test in end_test_data:
                response=requests.post(baseurl+str(i[1]),json=test)
            status_code_post=response.status_code
            if(status_code_post==200):
                print("route",str(i[1]),"post successful")
            else:
                print(i[1],"failed with return status_code",status_code_post)





if __name__ == '__main__':
    get_routes()



import fire
from .template import startproject
from .dockerise import createdockerfile, dockerfilechecker, buildimage, runimage
from .apitests import get_routes,tests
import os

def main():
  fire.Fire(TorchBlaze)

class TorchBlaze(object):
    """The CLI class. The class methods will act as CLI commands.
    """

    def __init__(self):
        pass

    def generate_template(self, project_name):
        """Generates the project template upon running the 
        "deploy generate_template --project_name <name>" command

        Arguments:
            project_name::str- Name of the project

        Returns:
            None
        """
        startproject(project_name)

    def generate_docker(self, image_name):
        """Generates the template Dockerfile and Docker image for the project.

        Arguments:
            image_name::str- Name of the Docker image that is to be generated.
        
        Returns:
            None
        """

        # Checking for Docker file exits or not 
        if not dockerfilechecker():
            # Creating a Docker file 
        	createdockerfile()
        	print('Default Dockerfile created.')
       	else:
            print('A Dockerfile already exists for this project.')
        # Building a Docker Image 
        buildimage(image_name)


    def run_docker(self, image_name):
        """Generates the template Dockerfile and Docker image for the project.

        Arguments:
            image_name::str- Name of the Docker image that is to be generated.
        
        Returns:
            None
        """
        runimage(image_name)


    def api_test(self):
        """A automated API testing used to check whether the API route is working correctly or not

        Arguments:
            None
        
        Returns:
            None
        """
        # Firstly we need to run the file and then start testing
        # Getting the list of routes from the app.py 
        routes=get_routes()
        # print(routes)
        HOST_ENDPOINT="http://127.0.0.1:8080"
        tests(routes,HOST_ENDPOINT)



if __name__ == '__main__':
    main()



import os
import pkg_resources
import signal
import sys

def signal_handler(sig, frame):
    print('You pressed Ctrl+C!')
    sys.exit(0)


def createdockerfile():
    """Creates a template Dockerfile for the project.
    
    Arguments:
        None
    
    Returns:
        None
    """
    curr_dir = os.getcwd()
    f = os.path.join(curr_dir, 'Dockerfile')
    with open(f, 'w+') as writefile:
        writefile.write(pkg_resources.resource_string('torchblaze',
                        'template_files/docker.txt').decode('utf-8'))


def dockerfilechecker():
    """Checks if a Dockerfile already exists in the project directory.

    Arguments:
        None
    Returns:
        Bool: True if the Dockerfile already exists, otherwise false. 
    """
    curr_dir = os.getcwd()
    rootfiles = os.listdir(curr_dir)
    dockerfilename = 'Dockerfile'
    if dockerfilename in rootfiles:
        return True
    return False


def buildimage(image_name:str):
    """Creates the Docker image for the API/project.

    Arguments:
        image_name::str- Name for the Docker image
    
    Returns:
        None
    """
    print('Docker Image ' + image_name + ' is getting created')
    os.system('docker build -t ' + image_name + ' .')
    print('Docker image build completed')


def runimage(image_name:str):
    """Runs the Docker Image Container

    Arguments:
        image_name::str- Name for the Docker image
    
    Returns:
        None
    """
    print("Docker Image Running")
    os.system("docker run -p 8080:8080 "+image_name)
    